/*
  file:   mod_auth_any.c
  author: Nafees Bin Zafar <binzafar@musc.edu>
  date:   7/12/00

  This used to be mod_auth.c in the Apache source tree.

  I warped it do validation by passing in the username and
  password into a popen()'d external program.

  Patch by robert (gorlitsk@musc.edu) to quote around password

  7/2002 - Robert Gorlitsky (gorlitsk@musc.edu)
           Added proper escaping and quoting of the password
*/

/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_auth: authentication
 * 
 * Rob McCool
 * 
 * Adapted to Apache by rst.
 *
 * dirkx - Added Authoritative control to allow passing on to lower
 *         modules if and only if the user-id is not known to this
 *         module. A known user with a faulty or absent password still
 *         causes an AuthRequired. The default is 'Authoritative', i.e.
 *         no control is passed along.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include <stdio.h>

/*
  mod_auth_any Mini-FAQ
  
  What is it?
  This is an Apache HTTP server runtime module.  Let's you use any external
  program as the user validator.
  
  What is being done in this mod?
  This mod gets the password given by the http client.  The script to be
  executed will be defined in the AuthUserfile directive in the .htaccess
  file or whatever the directive file is.

  How do we use it?
  Simply use the proper directives in the .htaccess file.

  What does that mean?
  There are certain configuration commands you can place in a Apache
  directory configuration file.  These commands are called directives.
  By default this file is called .htaccess. Depending on the "directives"
  you provide, Apache applies different modules to that directory.

  So what directives do I use?
  To be entirely unambiguous with the rest of the authentication modules
  I've introduced 3 new directives:
  AuthAnyUserProg
  AuthAnyGroupFile
  AuthAnyAuthoritative

  
*/

  
/* NB: debugging stuff */
extern int errno;


typedef struct auth_config_struct {
    char *auth_pwfile;
    char *auth_grpfile;
    int auth_authoritative;
} auth_config_rec;

static void *create_auth_dir_config(pool *p, char *d)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_pcalloc(p, sizeof(auth_config_rec));
    sec->auth_pwfile = NULL;	/* just to illustrate the default really */
    sec->auth_grpfile = NULL;	/* unless you have a broken HP cc */
    sec->auth_authoritative = 1;	/* keep the fortress secure by default */
    return sec;
}

static const char *set_auth_slot(cmd_parms *cmd, void *offset, char *f, char *t)
{
    if (t && strcmp(t, "standard"))
	return ap_pstrcat(cmd->pool, "Invalid auth file type: ", t, NULL);

    return ap_set_file_slot(cmd, offset, f);
}

static const command_rec auth_any_cmds[] =
{
    {"AuthAnyUserProg", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_pwfile), OR_AUTHCFG, TAKE12,
     "text file containing user IDs and passwords"},
    {"AuthAnyGroupFile", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_grpfile), OR_AUTHCFG, TAKE12,
     "text file containing group names and member user IDs"},
    {"AuthAnyAuthoritative", ap_set_flag_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_authoritative),
     OR_AUTHCFG, FLAG,
     "Set to 'off' to allow access control to be passed along to lower modules if the UserID is not known to this module"},
    {NULL}
};

module MODULE_VAR_EXPORT auth_any_module;

/* Escape special characters in the input string so that the bash
	shell will not interpolate them when the input string is withing
	single quotes.

   IN: null-terminated character array containing string to be interpolated
  OUT: newly allocate (using malloc) null-terminated character array
       containing the input string with the special characters properly
       escaped */
char* bash_single_quote_escape_string(const char *s) {
  /* used to count the length of the string and the number of single quotes */
  int str_len, sq_count;
  int s_pos, buf_pos; /* copy chars loop counter */
  /* used to hold the final result string */
  char *buf;

  /* Count the single quotes.
	  LOOP INVARIANT: str_len < (number of chars in string 's')
      POSTCONDITION: sq_count == (number of single quotes in string 's') */
  for (str_len = 0, sq_count = 0; s[str_len] != '\0'; str_len++)
	 if (s[str_len] == '\'')
		sq_count++;

  /* Allocate the memory for the final string.
	  Each ' (one char) will become '\'' (4 chars), so multiply by 4
     and don't forget to add 1 for terminating null. */
  buf = (char*) malloc(sizeof(char) * (str_len + 1 + sq_count * 4));

  /* Copy the chars of 's' into 'buf', turning each ' into '\'' */
  for (s_pos = 0, buf_pos = 0; s_pos < str_len; s_pos++) {
	 /* If we see a single quote, then put '\'' into 'buf' and advance
		 buf_pos 4 positions, else put the next char from 's' into 'buf'
		 and advance buf_pos 1 position. */
	 if(s[s_pos] == '\'') {
		buf[buf_pos++] = '\'';
		buf[buf_pos++] = '\\';
		buf[buf_pos++] = '\'';
		buf[buf_pos++] = '\'';
	 } else {
		buf[buf_pos++] = s[s_pos];
	 }
  }
  /* don't forget the null terminator */
  buf[buf_pos] = '\0';

  return buf;
}

/*
  nb:

  The popen() call to the external validator program is made here.
*/
static char *get_pw(request_rec *r, char *user, const char* password, char *auth_pwfile) {
    configfile_t *f;
    char* execstr;
    char* l;
    const char *rpw, *w;
    FILE* ext_authprog;
    FILE* fp;
	 char *escaped_password;
#ifdef __sun
    static char m_envstr[256]; /* sun's putenv() call is a bunch of bull malarkey */
#endif
    
    l = (char*) malloc (MAX_STRING_LEN * sizeof(char));
    execstr = (char*) malloc (MAX_STRING_LEN * sizeof(char));

#ifdef __sun
    snprintf (m_envstr, 256, "%s=%s\0", "REMOTE_ADDR", r -> connection -> remote_ip);
    putenv (m_envstr);
#else
    setenv ("REMOTE_ADDR", r -> connection -> remote_ip, 1);
#endif

    /* escape the password */
	 escaped_password = bash_single_quote_escape_string(password);

    /* open the program stream */
    snprintf (execstr, MAX_STRING_LEN, "%s %s \'%s\'", auth_pwfile, user, escaped_password);

	 /* free the escaped password before we forget */
	 free(escaped_password);

    if (!(ext_authprog = popen (execstr, "r"))) {

      ap_log_rerror (APLOG_MARK, APLOG_ERR, r, "Could not popen() on program: %s: %s",
		     auth_pwfile, strerror(errno));
      return NULL;
    }
    /* whew, the popen() has seemingly succeeded, let's write the username passwd stuff*/
    if (feof(ext_authprog) == 0)
      fgets (l, MAX_STRING_LEN, ext_authprog);
    pclose (ext_authprog);
    
    if (strncmp(l, "Authentication Error", 19) == 0) /* no authorized */
      return NULL;
    /* otherwise, we got a winner :), return user's real name */
    return l;

    /* ignore the rest */
    /*    
    if (!(f = ap_pcfg_openfile(r->pool, auth_pwfile))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
		    "Could not open password file: %s", auth_pwfile);
	return NULL;
    }
    
    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
	if ((l[0] == '#') || (!l[0]))
	    continue;
	rpw = l;
	w = ap_getword(r->pool, &rpw, ':');

	if (!strcmp(user, w)) {
	    ap_cfg_closefile(f);
	    return ap_getword(r->pool, &rpw, ':');
	}
    }
    ap_cfg_closefile(f);
    return NULL;
    */
}

static table *groups_for_user(pool *p, char *user, char *grpfile)
{
    configfile_t *f;
    table *grps = ap_make_table(p, 15);
    pool *sp;
    char l[MAX_STRING_LEN];
    const char *group_name, *ll, *w;

    if (!(f = ap_pcfg_openfile(p, grpfile))) {
/*add?	aplog_error(APLOG_MARK, APLOG_ERR, NULL,
		    "Could not open group file: %s", grpfile);*/
	return NULL;
    }

    sp = ap_make_sub_pool(p);

    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
	if ((l[0] == '#') || (!l[0]))
	    continue;
	ll = l;
	ap_clear_pool(sp);

	group_name = ap_getword(sp, &ll, ':');

	while (ll[0]) {
	    w = ap_getword_conf(sp, &ll);
	    if (!strcmp(w, user)) {
		ap_table_setn(grps, ap_pstrdup(p, group_name), "in");
		break;
	    }
	}
    }
    ap_cfg_closefile(f);
    ap_destroy_pool(sp);
    return grps;
}

/* These functions return 0 if client is OK, and proper error status
 * if not... either AUTH_REQUIRED, if we made a check, and it failed, or
 * SERVER_ERROR, if things are so totally confused that we couldn't
 * figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */

static int authenticate_basic_user(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_any_module);
    conn_rec *c = r->connection;
    const char *sent_pw;
    char *real_pw;
    char *invalid_pw;
    int res;
    FILE* fp;
    
    if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
	return res;

    if (!sec->auth_pwfile)
	return DECLINED;

    /*
      the follow get_pw() call needs to be modified to handle arbitrary
      sources for the password record.

      i'm assuming that ap_validate_password() does the actual authentication
      which is a bit of a problem since we're not necessarily maintaining
      the password in the same md5 or blowfish scheme used by apache.  and
      i don't really want to pass the entire crypt password token from the
      source.  right now i think disabling that call is our best recourse.

      Since we are relying on an external program to do validation, we
      probably can't distinguish between failure types (no such user,
      bad passwd, etc etc).  I should really handle success by evaluating
      the return code of the program, BUUUUUT I can't spend time on that right
      now.....
    */

    if (!(real_pw = get_pw(r, c->user, sent_pw, sec->auth_pwfile))) {
	if (!(sec->auth_authoritative))
	    return DECLINED;
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		      "user %s: Authentication failure from: %s URI: %s",
		      c->user, sec->auth_pwfile, r->uri);
	ap_note_basic_auth_failure(r);
	return AUTH_REQUIRED;
    }

    /*
      nb: eventually we will set some environment variable to reflect the token
          returned
    */
    return OK;
    
    /*
    invalid_pw = ap_validate_password(sent_pw, real_pw);
    
    if (invalid_pw != NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		      "user %s: authentication failure for \"%s\": %s",
		      c->user, r->uri, invalid_pw);
	ap_note_basic_auth_failure(r);
	return AUTH_REQUIRED;
    }

    return OK;
    */
}

/* Checking ID */

static int check_user_access(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_any_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    table *grpstatus;
    const array_header *reqs_arr = ap_requires(r);
    require_line *reqs;
    FILE* fp;

    
    /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive, 
     * then any user will do.
     */
    if (!reqs_arr)
	return (OK);
    reqs = (require_line *) reqs_arr->elts;

    if (sec->auth_grpfile)
	grpstatus = groups_for_user(r->pool, user, sec->auth_grpfile);
    else
	grpstatus = NULL;

    for (x = 0; x < reqs_arr->nelts; x++) {

	if (!(reqs[x].method_mask & (1 << m)))
	    continue;

	method_restricted = 1;

	t = reqs[x].requirement;
	w = ap_getword_white(r->pool, &t);
	if (!strcmp(w, "valid-user"))
	    return OK;
	if (!strcmp(w, "user")) {
	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (!strcmp(user, w))
		    return OK;
	    }
	}
	else if (!strcmp(w, "group")) {
	    if (!grpstatus)
		return DECLINED;	/* DBM group?  Something else? */

	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (ap_table_get(grpstatus, w))
		    return OK;
	    }
	} else if (sec->auth_authoritative) {
	    /* if we aren't authoritative, any require directive could be
	     * valid even if we don't grok it.  However, if we are 
	     * authoritative, we can warn the user they did something wrong.
	     * That something could be a missing "AuthAuthoritative off", but
	     * more likely is a typo in the require directive.
	     */
	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		"access to %s failed, reason: unknown require directive:"
		"\"%s\"", r->uri, reqs[x].requirement);
	}
    }

    if (!method_restricted)
	return OK;

    if (!(sec->auth_authoritative))
	return DECLINED;

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
	"access to %s failed, reason: user %s not allowed access",
	r->uri, user);
	
    ap_note_basic_auth_failure(r);
    return AUTH_REQUIRED;
}

module MODULE_VAR_EXPORT auth_any_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    create_auth_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    auth_any_cmds,		/* command table */
    NULL,			/* handlers */
    NULL,			/* filename translation */
    authenticate_basic_user,	/* check_user_id */
    check_user_access,		/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
